"""Common functions that are used and are quite helpful"""

from .util import (bindown, calculate_weight, class_for_name,
                   class_from_keyword, clip_native_to_wngrid,
                   compute_bin_edges, compute_dz, conversion_factor,
                   create_grid_res, decode_string_array, ensure_string_utf8,
                   find_closest_pair, get_molecular_weight, has_duplicates,
                   loadtxt2d, mass, merge_elements, molecule_texlabel,
                   movingaverage, quantile_corner, random_int_iter,
                   read_error_into_dict, read_error_line, read_table,
                   recursively_save_dict_contents_to_output,
                   sanitize_molecule_string, split_molecule_elements,
                   store_thing, tokenize_molecule, weighted_avg_and_std,
                   wnwidth_to_wlwidth)

__all__ = [
    "mass",
    "calculate_weight",
    "split_molecule_elements",
    "tokenize_molecule",
    "merge_elements",
    "split_molecule_elements",
    "sanitize_molecule_string",
    "get_molecular_weight",
    "molecule_texlabel",
    "bindown",
    "movingaverage",
    "quantile_corner",
    "loadtxt2d",
    "read_error_line",
    "read_error_into_dict",
    "read_table",
    "decode_string_array",
    "recursively_save_dict_contents_to_output",
    "store_thing",
    "weighted_avg_and_std",
    "random_int_iter",
    "compute_bin_edges",
    "clip_native_to_wngrid",
    "wnwidth_to_wlwidth",
    "class_from_keyword",
    "class_for_name",
    "create_grid_res",
    "conversion_factor",
    "compute_dz",
    "has_duplicates",
    "find_closest_pair",
    "ensure_string_utf8",
]
